@@ -1,25 +1,40 @@
+import os
+import re
+from subprocess import PIPE, Popen
 import sys
-from subprocess import Popen, PIPE
-
-from ..command import ls, validate
 
 from click.testing import CliRunner
 import pytest
-import os
+
+from ..command import __all_commands__, ls, validate
 
 
 @pytest.mark.parametrize("command", (ls, validate))
-def test_smoke(simple1_nwb, command):
+def test_smoke(organized_nwb_dir, command):
     runner = CliRunner()
-    r = runner.invoke(command, [simple1_nwb])
+    r = runner.invoke(command, [str(organized_nwb_dir)])
     assert r.exit_code == 0, f"Exited abnormally. out={r.stdout}"
     assert r.stdout, "There were no output whatsoever"
 
     # empty invocation should not crash
-    r = runner.invoke(command, [])
+    # But we must cd to the temp directory since current directory could
+    # have all kinds of files which could trip the command, e.g. validate
+    # could find some broken test files in the code base
+    with runner.isolated_filesystem():
+        r = runner.invoke(command, [])
     assert r.exit_code == 0, f"Exited abnormally. out={r.stdout}"
 
 
+@pytest.mark.parametrize("command", __all_commands__)
+def test_smoke_help(command):
+    runner = CliRunner()
+    r = runner.invoke(command, ["--help"])
+    assert r.exit_code == 0, f"Exited abnormally. out={r.stdout}"
+    assert r.stdout, "There were no output whatsoever"
+
+    assert re.match("Usage: .*Options:.*--help", r.stdout, flags=re.DOTALL) is not None
+
+
 def test_no_heavy_imports():
     # Timing --version for being fast is unreliable, so we will verify that
     # no h5py or numpy (just in case) module is imported upon import of the
@@ -31,9 +46,11 @@ def test_no_heavy_imports():
         [
             sys.executable,
             "-c",
-            "import sys; "
-            "import dandi.cli.command; "
-            "print(','.join(set(m.split('.')[0] for m in sys.modules)));",
+            (
+                "import sys; "
+                "import dandi.cli.command; "
+                "print(','.join(set(m.split('.')[0] for m in sys.modules)));"
+            ),
         ],
         env=env,
         stdout=PIPE,
@@ -44,5 +61,5 @@ def test_no_heavy_imports():
     loaded_heavy = set(modules).intersection(heavy_modules)
 
     assert not loaded_heavy
-    assert not stderr or b"dandi version" in stderr
+    assert not stderr or b"Failed to check" in stderr or b"dandi version" in stderr
     assert not p.wait()
