@@ -1,25 +1,32 @@
+from __future__ import annotations
+
 from collections import Counter
-from distutils.version import LooseVersion
 import os
 import os.path as op
+from pathlib import Path
 import re
+from typing import Any, Dict, List, Tuple, TypeVar, Union, cast
 import warnings
 
+import dandischema
 from fscacher import PersistentCache
 import h5py
 import hdmf
-import numpy as np
+from packaging.version import Version
 import pynwb
 from pynwb import NWBHDF5IO
 import semantic_version
 
 from . import __version__, get_logger
 from .consts import (
+    VIDEO_FILE_EXTENSIONS,
+    VIDEO_FILE_MODULES,
     metadata_nwb_computed_fields,
     metadata_nwb_file_fields,
     metadata_nwb_subject_fields,
 )
-from .utils import get_module_version
+from .utils import get_module_version, is_url
+from .validate_types import Scope, Severity, ValidationOrigin, ValidationResult
 
 lgr = get_logger()
 
@@ -35,7 +42,9 @@ metadata_cache = PersistentCache(
     name="dandi-metadata", tokens=dandi_cache_tokens, envvar="DANDI_CACHE"
 )
 validate_cache = PersistentCache(
-    name="dandi-validate", tokens=dandi_cache_tokens, envvar="DANDI_CACHE"
+    name="dandi-validate",
+    tokens=dandi_cache_tokens + [get_module_version(dandischema)],
+    envvar="DANDI_CACHE",
 )
 
 
@@ -50,7 +59,7 @@ def _sanitize_nwb_version(v, filename=None, log=None):
         log = lgr.warning
     elif not log:
 
-        def log(v):  # does nothing
+        def log(v: str) -> str:  # does nothing
             return v
 
     if isinstance(v, str):
@@ -70,8 +79,7 @@ def _sanitize_nwb_version(v, filename=None, log=None):
         log(f"{msg} is not text which follows semver specification")
 
     if isinstance(v, str) and not semantic_version.validate(v):
-        msgtype = "error" if LooseVersion(v) >= "2.1.0" else "warning"
-        log(f"{msgtype}: {msg} is not a proper semantic version. See http://semver.org")
+        log(f"error: {msg} is not a proper semantic version. See http://semver.org")
 
     return v
 
@@ -103,10 +111,11 @@ def get_nwb_version(filepath, sanitize=False):
         try:
             return _sanitize(h5file["nwb_version"][...].tostring().decode())
         except Exception:
-            lgr.debug("%s has no nwb_version" % filepath)
+            lgr.debug("%s has no nwb_version", filepath)
+    return None
 
 
-def get_neurodata_types_to_modalities_map():
+def get_neurodata_types_to_modalities_map() -> Dict[str, str]:
     """Return a dict to map neurodata types known to pynwb to "modalities"
 
     It is an ugly hack, largely to check feasibility.
@@ -115,7 +124,7 @@ def get_neurodata_types_to_modalities_map():
     """
     import inspect
 
-    ndtypes = {}
+    ndtypes: Dict[str, str] = {}
 
     # TODO: if there are extensions, they might have types subclassed from the base
     # types.  There might be a map within pynwb (pynwb.get_type_map?) to return
@@ -150,7 +159,7 @@ def get_neurodata_types_to_modalities_map():
 
 
 @metadata_cache.memoize_path
-def get_neurodata_types(filepath):
+def get_neurodata_types(filepath: Union[str, Path]) -> List[str]:
     with h5py.File(filepath, "r") as h5file:
         all_pairs = _scan_neurodata_types(h5file)
 
@@ -168,7 +177,7 @@ def get_neurodata_types(filepath):
     return out
 
 
-def _scan_neurodata_types(grp):
+def _scan_neurodata_types(grp: h5py.File) -> List[Tuple[Any, Any]]:
     out = []
     if "neurodata_type" in grp.attrs:
         out.append((grp.attrs["neurodata_type"], grp.attrs.get("description", None)))
@@ -178,7 +187,7 @@ def _scan_neurodata_types(grp):
     return out
 
 
-def _get_pynwb_metadata(path):
+def _get_pynwb_metadata(path: Union[str, Path]) -> Dict[str, Any]:
     out = {}
     with NWBHDF5IO(path, "r", load_namespaces=True) as io:
         nwb = io.read()
@@ -208,7 +217,7 @@ def _get_pynwb_metadata(path):
             out.update(dandi_icephys.fields)
         # Go through devices and see if there any probes used to record this file
         probe_ids = [
-            np.asscalar(v.probe_id)  # .asscalar to avoid numpy types
+            v.probe_id.item()  # .item to avoid numpy types
             for v in getattr(nwb, "devices", {}).values()
             if hasattr(v, "probe_id")  # duck typing
         ]
@@ -226,11 +235,97 @@ def _get_pynwb_metadata(path):
             key = f[len("number_of_") :]
             out[f] = len(getattr(nwb, key, []) or [])
 
+        # get external_file data:
+        out["external_file_objects"] = _get_image_series(nwb)
+
     return out
 
 
+def _get_image_series(nwb: pynwb.NWBFile) -> List[dict]:
+    """Retrieves all ImageSeries related metadata from an open nwb file.
+
+    Specifically it pulls out the ImageSeries uuid, name and all the
+    externally linked files named under the argument 'external_file'.
+
+    Parameters
+    ----------
+    nwb: pynwb.NWBFile
+
+    Returns
+    -------
+    out: List[dict]
+        list of dicts : [{id: <ImageSeries uuid>, name: <IMageSeries name>,
+        external_files=[ImageSeries.external_file]}]
+        if no ImageSeries found in the given modules to check, then it returns an empty list.
+    """
+    out = []
+    for module_name in VIDEO_FILE_MODULES:
+        module_cont = getattr(nwb, module_name)
+        for name, ob in module_cont.items():
+            if isinstance(ob, pynwb.image.ImageSeries) and ob.external_file is not None:
+                out_dict = dict(id=ob.object_id, name=ob.name, external_files=[])
+                for ext_file in ob.external_file:
+                    if Path(ext_file).suffix in VIDEO_FILE_EXTENSIONS:
+                        out_dict["external_files"].append(Path(ext_file))
+                    else:
+                        lgr.warning(
+                            "external file %s should be one of: %s",
+                            ext_file,
+                            ", ".join(VIDEO_FILE_EXTENSIONS),
+                        )
+                out.append(out_dict)
+    return out
+
+
+def rename_nwb_external_files(metadata: List[dict], dandiset_path: str) -> None:
+    """Renames the external_file attribute in an ImageSeries datatype in an open nwb file.
+
+    It pulls information about the ImageSeries objects from metadata:
+    metadata["external_file_objects"] populated during _get_pynwb_metadata() call.
+
+    Parameters
+    ----------
+    metadata: List[dict]
+        list of dictionaries containing the metadata gathered from the nwbfile
+    dandiset_path: str
+        base path of dandiset
+    """
+    for meta in metadata:
+        if not all(i in meta for i in ["path", "dandi_path", "external_file_objects"]):
+            lgr.warning(
+                "could not rename external files, update metadata "
+                'with "path", "dandi_path", "external_file_objects"'
+            )
+            return
+        dandiset_nwbfile_path = op.join(dandiset_path, meta["dandi_path"])
+        with NWBHDF5IO(dandiset_nwbfile_path, mode="r+", load_namespaces=True) as io:
+            nwb = io.read()
+            for ext_file_dict in meta["external_file_objects"]:
+                # retrieve nwb neurodata object of the given object id:
+                container_list = [
+                    child
+                    for child in nwb.children
+                    if ext_file_dict["id"] == child.object_id
+                ]
+                if len(container_list) == 0:
+                    continue
+                else:
+                    container = container_list[0]
+                # rename all external files:
+                for no, (name_old, name_new) in enumerate(
+                    zip(
+                        ext_file_dict["external_files"],
+                        ext_file_dict["external_files_renamed"],
+                    )
+                ):
+                    if not is_url(str(name_old)):
+                        container.external_file[no] = str(name_new)
+
+
 @validate_cache.memoize_path
-def validate(path, devel_debug=False):
+def validate(
+    path: Union[str, Path], devel_debug: bool = False
+) -> List[ValidationResult]:
     """Run validation on a file and return errors
 
     In case of an exception being thrown, an error message added to the
@@ -241,20 +336,49 @@ def validate(path, devel_debug=False):
     path: str or Path
     """
     path = str(path)  # Might come in as pathlib's PATH
+    errors: List[ValidationResult] = []
     try:
-        with pynwb.NWBHDF5IO(path, "r", load_namespaces=True) as reader:
-            errors = pynwb.validate(reader)
-        lgr.warning(
-            "pynwb validation errors for %s: %s",
-            path,
-            errors,
-            extra={"validating": True},
-        )
+        if Version(pynwb.__version__) >= Version(
+            "2.2.0"
+        ):  # Use cached namespace feature
+            # argument get_cached_namespaces is True by default
+            error_outputs, _ = pynwb.validate(paths=[path])
+        else:  # Fallback if an older version
+            with pynwb.NWBHDF5IO(path=path, mode="r", load_namespaces=True) as reader:
+                error_outputs = pynwb.validate(io=reader)
+        # TODO: return ValidationResult structs
+        for error_output in error_outputs:
+            errors.append(
+                ValidationResult(
+                    origin=ValidationOrigin(
+                        name="pynwb",
+                        version=pynwb.__version__,
+                    ),
+                    severity=Severity.WARNING,
+                    id=f"pywnb.{error_output}",
+                    scope=Scope.FILE,
+                    path=Path(path),
+                    message="Failed to validate.",
+                )
+            )
     except Exception as exc:
         if devel_debug:
             raise
-        lgr.warning("Failed to validate %s: %s", path, exc, extra={"validating": True})
-        errors = [f"Failed to validate {path}: {exc}"]
+        errors.extend(
+            [
+                ValidationResult(
+                    origin=ValidationOrigin(
+                        name="pynwb",
+                        version=pynwb.__version__,
+                    ),
+                    severity=Severity.ERROR,
+                    id="pywnb.GENERIC",
+                    scope=Scope.FILE,
+                    path=Path(path),
+                    message=f"{exc}",
+                )
+            ]
+        )
 
     # To overcome
     #   https://github.com/NeurodataWithoutBorders/pynwb/issues/1090
@@ -271,12 +395,35 @@ def validate(path, devel_debug=False):
     else:
         if version is not None:
             # Explicitly sanitize so we collect warnings.
-            # TODO: later cast into proper ERRORs
-            version = _sanitize_nwb_version(version, log=errors.append)
-            loosever = LooseVersion(version)
-            if loosever and loosever < "2.1.0":
+            nwb_errors: list[str] = []
+            version = _sanitize_nwb_version(version, log=nwb_errors.append)
+            for e in nwb_errors:
+                errors.append(
+                    ValidationResult(
+                        origin=ValidationOrigin(
+                            name="pynwb",
+                            version=pynwb.__version__,
+                        ),
+                        severity=Severity.ERROR,
+                        id="pywnb.GENERIC",
+                        scope=Scope.FILE,
+                        path=Path(path),
+                        message=e,
+                    )
+                )
+            # Do we really need this custom internal function? string comparison works fine.
+            try:
+                v = semantic_version.Version(version)
+            except ValueError:
+                v = None
+            if v is not None and v < semantic_version.Version("2.1.0"):
                 errors_ = errors[:]
-                errors = [e for e in errors if not re_ok_prior_210.search(str(e))]
+                errors = [
+                    e
+                    for e in errors
+                    if not re_ok_prior_210.search(cast(str, getattr(e, "message", "")))
+                ]
+                # This is not an error, just logging about the process, hence logging:
                 if errors != errors_:
                     lgr.debug(
                         "Filtered out %d validation errors on %s",
@@ -295,7 +442,7 @@ def validate(path, devel_debug=False):
 _ignored_benign_pynwb_warnings = False
 
 
-def ignore_benign_pynwb_warnings():
+def ignore_benign_pynwb_warnings() -> None:
     global _ignored_benign_pynwb_warnings
     if _ignored_benign_pynwb_warnings:
         return
@@ -308,7 +455,7 @@ def ignore_benign_pynwb_warnings():
     _ignored_benign_pynwb_warnings = True
 
 
-def get_object_id(path):
+def get_object_id(path: Union[str, Path]) -> Any:
     """Read, if present an object_id
 
     if not available -- would simply raise a corresponding exception
@@ -317,7 +464,12 @@ def get_object_id(path):
         return f.attrs["object_id"]
 
 
-def make_nwb_file(filename, *args, cache_spec=False, **kwargs):
+StrPath = TypeVar("StrPath", str, Path)
+
+
+def make_nwb_file(
+    filename: StrPath, *args: Any, cache_spec: bool = False, **kwargs: Any
+) -> StrPath:
     """A little helper to produce an .nwb file in the path using NWBFile
 
     Note: it doesn't cache_spec by default
@@ -328,7 +480,7 @@ def make_nwb_file(filename, *args, cache_spec=False, **kwargs):
     return filename
 
 
-def copy_nwb_file(src, dest):
+def copy_nwb_file(src: Union[str, Path], dest: Union[str, Path]) -> str:
     """ "Copy" .nwb file by opening and saving into a new path.
 
     New file (`dest`) then should have new `object_id` attribute, and thus be
@@ -352,9 +504,32 @@ def copy_nwb_file(src, dest):
         dest = op.join(dest, op.basename(src))
     else:
         os.makedirs(op.dirname(dest), exist_ok=True)
-    # The simplest way yoh could find
-    with pynwb.NWBHDF5IO(src, "r") as ior, pynwb.NWBHDF5IO(
-        dest, "w", manager=ior.manager
-    ) as iow:
-        iow.write(ior.read().copy(), link_data=False)
-    return dest
+    with pynwb.NWBHDF5IO(src, "r") as ior, pynwb.NWBHDF5IO(dest, "w") as iow:
+        data = ior.read()
+        data.generate_new_id()
+        iow.export(ior, nwbfile=data)
+    return str(dest)
+
+
+@metadata_cache.memoize_path
+def nwb_has_external_links(filepath: Union[str, Path]) -> bool:
+    with h5py.File(filepath, "r") as fp:
+        visited = set()
+
+        # cannot use `file.visititems` because it skips external links
+        # (https://github.com/h5py/h5py/issues/671)
+        def visit(path: str = "/") -> bool:
+            if isinstance(fp[path], h5py.Group):
+                for key in fp[path].keys():
+                    key_path = path + "/" + key
+                    if key_path not in visited:
+                        visited.add(key_path)
+                        if isinstance(
+                            fp.get(key_path, getlink=True), h5py.ExternalLink
+                        ) or visit(key_path):
+                            return True
+            elif isinstance(fp.get(path, getlink=True), h5py.ExternalLink):
+                return True
+            return False
+
+        return visit()
